/*****************************************************************************
 * This is the cfht_dph2psf source code in C language.
 * This program does the same thing as the IDL program cfht_dph2psf.pro, that
 * is it allows to compute PUEO's PSF associated with an object
 * from the various dph (structure function) files saved during the
 * acquisition and associated image of point sources.
 * cfht_dph2psf take the name of a single parameter file as input.
 *  for example: cfht_dph2psf myparfile.par 
 * The syntax of this paramater file must be the following:
 * * PSFILE
 * xxxxxxx        <- Name of the output file for the calculated PSF
 * 
 * * OBJECT
 * xxxxxxx        <- Name of the first dph file for the object
 * ......         <- Name of other dph files for the object (optional)
 *
 * * REFERENCE 1
 * xxxxxxx        <- Name of the image of the first reference
 * xxxxxxx        <- Name of the first dph file for the first reference
 * .......        <- Name of other dph files for the first reference (optional)
 *
 * .......
 * 
 * * REFERENCE n
 * xxxxxxx        <- Name of the image of the nth reference
 * xxxxxxx        <- Name of the first dph file for the nth reference
 * .......        <- Name of other dph files for the nth reference (optional) 
 *
 *
 * To compile, this source must be linked to the fitsio and math library
 * (see README for details).
 * --------------------------------------------------------------------------
 *
 * Version 1.0    11 March 1997  Jean-Pierre Veran (veran@hplyot.obspm.fr)
 *
 ****************************************************************************/
#include<stdio.h>
#include<string.h>
#include<math.h>
#include<errno.h>
#include "fitsio.h"

#define X_SIZE 128               /* X_SIZE of all arrays */
#define Y_SIZE 128               /* Y_SIZE of all arrays */
#define SIZE_ARRAY X_SIZE*Y_SIZE /* Number of pixels of all arrays */

static void aob_dph2psfcomp(float *,float *,float *,float *);
static void read_fits_file(char *,float *, float *, float *,int);
static char *find_next_keyword(char *,FILE *);
static char *get_line(char *,FILE *);
static void fourn(float [],int [],int,int);
static void fitsio_printerror( int);
static char *mybasename(char *);

/****************************************************************************
 * Main program.
 * Returns 0 upon successful completion.
 ****************************************************************************/
int main(int argc, char *argv[])
{
  char *filename;
  char line[255];
  FILE *fich;
  float ref[SIZE_ARRAY],psf[SIZE_ARRAY],dphref[SIZE_ARRAY],dphobj[SIZE_ARRAY];
  float tmp[SIZE_ARRAY];
  int i,n,nref;
  fitsfile *fptr;       /* pointer to the FITS file, defined in fitsio.h */  
  int status;
  float lambda, camsamp;
  char card[80],comment[80];
  long naxes[2] = {X_SIZE,Y_SIZE};

  if (argc != 2) {
    fprintf(stderr,"Usage: %s parameter_file\n",argv[0]);
    exit(1);
  }
  filename = argv[1];

  if((fich = fopen(filename,"r"))==NULL) {
    perror("ERROR: Unable to open parameter file for reading");
    exit(1);
  }

  /* First keyword should be * PSFILE, giving the name of the output file */
  if (strcmp(find_next_keyword(line,fich),"* PSFFIL")) {
    fprintf(stderr,"ERROR: Keyword * PSFFILE not found in %s\n",filename);
    exit(1);
  }
  if (strcmp(get_line(line,fich),"") ==0) {
    fprintf(stderr,"ERROR: name of PSF file is not found\n");
    exit(1);
  }

  status = 0;         /* initialize status before calling fitsio routines */

  /* Open output file for FITS writing */
  remove(line);               /* Delete old file if it already exists */
  if (fits_create_file(&fptr, line, &status)) /* create new FITS file */
    fitsio_printerror( status );

  /* write the required keywords for the primary array image */
  if ( fits_create_img(fptr,  -32, 2, naxes, &status) )
    fitsio_printerror( status );

  /* write DATE keyword */
  if ( fits_write_date (fptr, &status) )
    fitsio_printerror( status );

  lambda = 0 ; camsamp = 0;   /* They will be read from FITS header */

  /* Second keyword should be * OBJECT, giving the name of the dph for 
     the object. */
  if (strcmp(find_next_keyword(line,fich),"* OBJECT")) {
    fprintf(stderr,"ERROR: Keyword * OBJECT not found in %s\n",filename);
    exit(1);
  }
  for(i=0;i<SIZE_ARRAY;i++) dphobj[i] = (float)0;
  n = 0;
  while (strcmp(get_line(line,fich),"") !=0) {
    read_fits_file(line,tmp,&lambda,&camsamp,1);
    for(i=0;i<SIZE_ARRAY;i++) dphobj[i]=dphobj[i]+tmp[i];
    if (n == 0) {
      /* Write the LAMBDA and CAMSAMP cards */
      if ( fits_write_key(fptr,  TFLOAT, "LAMBDA", &lambda,
			  "Central imaging wavelength", &status) )
	fitsio_printerror( status );
      if ( fits_write_key(fptr, TFLOAT, "CAMSAMP", &camsamp,
			  "Size of one pixel (in arcsec)", &status) )
	fitsio_printerror( status );
    }
    /* Write name of current dph into FITS header */
    (void)sprintf(card,"DPHOBJ%d",n+1);
    (void)sprintf(comment,"Structure function %d for object",n+1);
    if ( fits_write_key(fptr, TSTRING, card, mybasename(line),
			  comment, &status) )
	fitsio_printerror( status );

    /* Iterate */
    n++;
  }
  if (n == 0) {
    fprintf(stderr,
	    "ERROR: at least one dph file must be given for the object\n");
    exit(1);
  }
  if (n > 1) for(i=0;i<SIZE_ARRAY;i++) dphobj[i]=dphobj[i]/n;

  /* Now we parse the description of the reference */
  nref = 0;
  while(strcmp(find_next_keyword(line,fich),"* REFERE")==0) {
    if (strcmp(get_line(line,fich),"") ==0) {
      fprintf(stderr,"ERROR: File name for reference image %d is not found\n",nref+1);
      exit(1);
    }
    read_fits_file(line,ref,NULL,NULL,0);

    /* Write name of current reference image in the header */
    (void)sprintf(card,"REF%d",nref+1);
    (void)sprintf(comment,"Reference source number %d",nref+1);
    if ( fits_write_key(fptr, TSTRING, card, mybasename(line),
			  comment, &status) )
	fitsio_printerror( status );

    /* Now look for the name of the corresponding dph files */
    for(i=0;i<SIZE_ARRAY;i++) dphref[i]=(float)0;
    n = 0;
    while (strcmp(get_line(line,fich),"") != 0) {
      read_fits_file(line,tmp,&lambda,&camsamp,1);
      for(i=0;i<SIZE_ARRAY;i++) dphref[i]=dphref[i]+tmp[i];

      /* Write name of current reference dph file in the header */
      (void)sprintf(card,"DREF%d_%d",nref+1,n+1);
      (void)sprintf(comment,"Structure function %d for reference %d",
		    n+1,nref+1);
      if ( fits_write_key(fptr, TSTRING, card, mybasename(line),
			  comment, &status) )
      fitsio_printerror( status );

      /* Iterate: next dph if any */
      n++;
    }
    if (n == 0) {
      fprintf(stderr,
	      "ERROR: at least one dph file must be given for each dph\n");
      exit(1);
    }
    if (n > 1) for(i=0;i<SIZE_ARRAY;i++) dphref[i]=dphref[i]/n;
    aob_dph2psfcomp(tmp,dphobj,dphref,ref);
    for(i=0;i<SIZE_ARRAY;i++) psf[i]=psf[i]+tmp[i];
    
    /* Iterate: next reference if any */
    nref  ++;
  }
  if (nref ==0) { 
    fprintf(stderr,
	      "ERROR: at least one reference must be used\n");
    exit(1);
  }
  if (nref > 1) for(i=0;i<SIZE_ARRAY;i++) psf[i]=psf[i]/nref;  

  /* write the PSF array to the FITS file */
  if ( fits_write_img(fptr, TFLOAT, (long)1, SIZE_ARRAY, psf, &status) )
    fitsio_printerror( status );            

  if ( fits_close_file(fptr, &status) )                /* close the file */
    fitsio_printerror( status );           
  
  /*printf("%lf %lf %lf %lf\n",psf[0],psf[1],psf[100*128+3],psf[64*128+64]);*/

  return(0);
}

/*****************************************************************************
 * This routine actually computes the PSF knowing the atmospheric structure
 * function for the reference dphref, an image of the reference ref and
 * the atmospheric structure function of the object.
 *****************************************************************************/
void aob_dph2psfcomp(float *psf,float *dphobj,float *dphref,float *ref)
{
  float otffilt[SIZE_ARRAY];
  float complexarr[2*SIZE_ARRAY];
  int i,i2;
  int nn[2] = {X_SIZE,Y_SIZE};
  double total;

  /*printf("Now computing PSF\n");*/

  /* Compute OTF of filter from difference of structure functions */
  for(i=0;i<SIZE_ARRAY;i++) otffilt[i] = exp(-0.5*(dphobj[i]-dphref[i]));

  /* Mask OTF: All values at 0 in dph must be 0 in OTF except at origin */
  for(i=1;i<SIZE_ARRAY;i++) if (dphobj[i] == 0) otffilt[i] = 0; 

  /* Compute OTF of reference image */
  for(i=0;i<SIZE_ARRAY;i++) {
    i2 = 2*i;
    complexarr[i2] = (float)ref[i];
    complexarr[i2+1] = (float)0;
  }
  fourn(complexarr-1,nn-1,2,1);

  /* Multiply by OTF of filter */
  for(i=0;i<SIZE_ARRAY;i++) {
    i2 = 2*i;
    complexarr[i2] = complexarr[i2] * otffilt[i];
    complexarr[i2+1] = complexarr[i2+1] * otffilt[i];
  }

  /* Compute PSF */
  fourn(complexarr-1,nn-1,2,-1);
  for(i=0;i<SIZE_ARRAY;i++) psf[i] = complexarr[2*i];

  /* Normalize */
  total = (double)0;
  for(i=0;i<SIZE_ARRAY;i++) total = total + psf[i];
  for(i=0;i<SIZE_ARRAY;i++) psf[i] = psf[i] / total;
}  

/****************************************************************************
 * Read the FITS format file filename and put image data in the mono-
 * dimensional floating point array a.
 * Also check that the dimension are X_SIZE by Y_SIZE.
 * Also if check is set and *lambda is not zero then check that the LAMBDA
 * in the header is equal to lambda. Same for camsamp and CAMSAMP.
 ****************************************************************************/
void read_fits_file(char *filename,float *a, float *lambda, float *camsamp,int check)
{
  fitsfile *fptr;
  int status, nfound, anynull;
  float nullval;
  long naxes[2];
  float gotlambda,gotcamsamp;

  /*printf("Now reading %s+\n",filename);*/
  status = 0;
  if ( fits_open_file(&fptr, filename, READONLY, &status) )
    fitsio_printerror( status );

  /* read the NAXIS1 and NAXIS2 keyword to get image size */
  if ( fits_read_keys_lng(fptr, "NAXIS", 1, 2, naxes, &nfound, &status) )
    fitsio_printerror( status );
  if ((naxes[0] != X_SIZE) || (naxes[1] != Y_SIZE)) {
    fprintf(stderr,
	    "ERROR: file %s is not a %dx%d image",filename,X_SIZE,Y_SIZE);
    exit(1);
  }  

  if (check) {
    /* read the LAMBDA and CAMSAMP keyword and check */
    float gotlambda, gotcamsamp;
    if ( fits_read_key(fptr, TFLOAT, "LAMBDA", &gotlambda, NULL, &status) )
      fitsio_printerror( status );
    if ( fits_read_key(fptr, TFLOAT, "CAMSAMP", &gotcamsamp, NULL, &status) )
      fitsio_printerror( status );
    if (*lambda != 0) {
      if (*lambda != gotlambda) {
	fprintf(stderr,
	 "ERROR: The LAMBDA keyword must be the same for all the dph files\n");
	exit(1);
      }
      if (*camsamp != gotcamsamp) {
	fprintf(stderr,
	"ERROR: The CAMSAMP keyword must be the same for all the dph files\n");
	exit(1);
      }
    }
    *lambda = gotlambda;
    *camsamp = gotcamsamp;
  }

  /* read-in the data */
  nullval = 0;
  if ( fits_read_img(fptr, TFLOAT, (long)1, SIZE_ARRAY, &nullval,
                  a, &anynull, &status) )
           fitsio_printerror( status );

    if ( fits_close_file(fptr, &status) )
         fitsio_printerror( status );

    return;
}

/*****************************************************/
/* Print out cfitsio error messages and exit program */
/* Copied from cookbook.c in cfitsio library.        */
/*****************************************************/
void fitsio_printerror( int status)
{
  char status_str[FLEN_STATUS], errmsg[FLEN_ERRMSG];
  
  if (status)
    fprintf(stderr, "\n*** Error occurred during FITS I/O ***\n");
  
  fits_get_errstatus(status, status_str);   /* get the error description */
  fprintf(stderr, "\nstatus = %d: %s\n", status, status_str);
  
  /* get first message; null if stack is empty */
  if ( fits_read_errmsg(errmsg) ) 
    {
      fprintf(stderr, "\nError message stack:\n");
      fprintf(stderr, " %s\n", errmsg);
      
      while ( fits_read_errmsg(errmsg) )  /* get remaining messages */
	fprintf(stderr, " %s\n", errmsg);
    }
  
  exit( status );       /* terminate the program, returning error status */
}

/****************************************************************************
 * This function is the equivalent of the basename Unix facility.
 ****************************************************************************/
char *mybasename(char *name)
{
  char *pt;

  pt = strrchr(name,'/'); /* pt points on last / charcacter or NULL if no / */
  if (pt != NULL) return(pt+1);
  return(name);
}

/****************************************************************************
 * Return next keyword value in parameter file.
 * A keyword is a string starting with "* ".
 * Return empty string if EOF is reached before finding any keyword.
 ****************************************************************************/
char *find_next_keyword(char *line,FILE *fich)
{
  while (!feof(fich)) {
    (void)get_line(line,fich);
    line[8] = '\0';   /* limit keyword scope to 8 characters */
    if (!strncmp(line,"* ",2)) return(line);
  }
  return(line); /* line has been set to empty string by get_line */
}

/****************************************************************************
 * Return the content of the current line in parameter file
 * If EOF is reached, return empty string
 ****************************************************************************/
char *get_line(char *line,FILE *fich)
{
  if(fgets(line,255,fich)!=NULL) {
    /* Normal case: we did read something on the line that we return */
    /* However, since fgets reads the \n, we must remove it */
    line[strlen(line)-1] = '\0';
    return(line);
  }
  /* Nothing could be read. What happened ? */
  if (feof(fich)) {
    /* That's all right: we just hit end-of-file. */
    /* We just return an empty string */
    line[0] = '\0';
    return(line);
  }
  /* Here we really have a problem */
  fprintf(stderr,"ERROR: Error while reading parameter file\n");
  exit(1);
}

/****************************************************************************
 * This is the multi-dimensional FFT program from Numerical Recipes in C.
 * data is a complex array (alternating real and imaginary part).
 * nn is the array containing the length of each dimension.
 * ndim is the number of dimension.
 * isign is 1 for the forward transform and -1 for the backward transform.
 * As a result, data is replaced by its FFT.
 * NB: Remember that as for all programs in Numerical Recipes, all arrays
 *     start with element number 1 instead of the usual 0.
 ****************************************************************************/
#define SWAP(a,b) tempr=(a);(a)=(b);(b)=tempr
void fourn(float data[],int nn[],int ndim,int isign)
{
	int i1,i2,i3,i2rev,i3rev,ip1,ip2,ip3,ifp1,ifp2;
	int ibit,idim,k1,k2,n,nprev,nrem,ntot;
	float tempi,tempr;
	double theta,wi,wpi,wpr,wr,wtemp;

	ntot=1;
	for (idim=1;idim<=ndim;idim++)
		ntot *= nn[idim];
	nprev=1;
	for (idim=ndim;idim>=1;idim--) {
		n=nn[idim];
		nrem=ntot/(n*nprev);
		ip1=nprev << 1;
		ip2=ip1*n;
		ip3=ip2*nrem;
		i2rev=1;
		for (i2=1;i2<=ip2;i2+=ip1) {
			if (i2 < i2rev) {
				for (i1=i2;i1<=i2+ip1-2;i1+=2) {
					for (i3=i1;i3<=ip3;i3+=ip2) {
						i3rev=i2rev+i3-i2;
						SWAP(data[i3],data[i3rev]);
						SWAP(data[i3+1],data[i3rev+1]);
					}
				}
			}
			ibit=ip2 >> 1;
			while (ibit >= ip1 && i2rev > ibit) {
				i2rev -= ibit;
				ibit >>= 1;
			}
			i2rev += ibit;
		}
		ifp1=ip1;
		while (ifp1 < ip2) {
			ifp2=ifp1 << 1;
			theta=isign*6.28318530717959/(ifp2/ip1);
			wtemp=sin(0.5*theta);
			wpr = -2.0*wtemp*wtemp;
			wpi=sin(theta);
			wr=1.0;
			wi=0.0;
			for (i3=1;i3<=ifp1;i3+=ip1) {
				for (i1=i3;i1<=i3+ip1-2;i1+=2) {
					for (i2=i1;i2<=ip3;i2+=ifp2) {
						k1=i2;
						k2=k1+ifp1;
						tempr=wr*data[k2]-wi*data[k2+1];
						tempi=wr*data[k2+1]+wi*data[k2];
						data[k2]=data[k1]-tempr;
						data[k2+1]=data[k1+1]-tempi;
						data[k1] += tempr;
						data[k1+1] += tempi;
					}
				}
				wr=(wtemp=wr)*wpr-wi*wpi+wr;
				wi=wi*wpr+wtemp*wpi+wi;
			}
			ifp1=ifp2;
		}
		nprev *= n;
	}
}

#undef SWAP



